from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _

from iso4217 import Currency

# TODO: Validate multiple msrp lines
def validate_msrp_format(value):
    """
    This is a validator function used to validate an MSRP value added to the Product.
    """
    msrp = []
    if not value:
        return None

    msrp = value.split(',')

    if len(msrp) != 2 or not msrp:
        raise ValidationError(_("Invalid MSRP Value"), params={'value': value})
    
    if not msrp[0] or not msrp[1]:
        raise ValidationError(_("Invalid MSRP Value"), params={'value': value})

    if not msrp[0].lower() in Currency.__dict__:
        raise ValidationError(_("Invalid MSRP Value"), params={'value': value})



"""
Base Payment processor used by all derived processors.
"""
import django.dispatch

from copy import deepcopy
from datetime import timedelta
from django.db.models import Sum
from django.conf import settings
from django.utils import timezone
from vendor.models import Payment, Invoice, Receipt
from vendor.models.choice import PurchaseStatus, TermType
##########
# SIGNALS

vendor_pre_authorization = django.dispatch.Signal()
vendor_process_payment =  django.dispatch.Signal()
vendor_post_authorization =  django.dispatch.Signal()

#############
# BASE CLASS

class PaymentProcessorBase(object):
    """
    Setup the core functionality for all processors.
    """
    status = None
    invoice = None
    provider = None
    payment = None
    payment_info = {}
    billing_address = {}
    transaction_token = None
    transaction_submitted = False
    transaction_message = {}
    transaction_response = {}


    def __init__(self, invoice):
        """
        This should not be overriden.  Override one of the methods it calls if you need to.
        """
        self.set_invoice(invoice)
        self.provider = self.__class__.__name__
        self.processor_setup()

    def processor_setup(self):
        """
        This is for setting up any of the settings needed for the payment processing.
        For example, here you would set the 
        """
        pass

    def set_payment_info(self, **kwargs):
        self.payment_info = kwargs

    def set_invoice(self, invoice):
        self.invoice = invoice

    def create_payment_model(self):
        self.payment = Payment(  profile=self.invoice.profile,
                            amount=self.invoice.total,
                            provider=self.provider,
                            invoice=self.invoice,
                            created=timezone.now()
                            )

    def save_payment_transaction(self):
        pass

    def update_invoice_status(self, new_status):
        if self.transaction_submitted:
            self.invoice.status = new_status
        else:
            self.invoice.status = Invoice.InvoiceStatus.CART
        self.invoice.save()

    def create_receipt_by_term_type(self, product, order_item, term_type):
        receipt = Receipt()
        receipt.profile = self.invoice.profile
        receipt.order_item = order_item
        receipt.transaction = self.payment.transaction
        receipt.status = PurchaseStatus.COMPLETE
        receipt.start_date = timezone.now()
        if term_type == TermType.SUBSCRIPTION:
            total_months = int(order_item.offer.term_details['period_length']) * int(order_item.offer.term_details['payment_occurrences'])
            receipt.end_date = timezone.now() + timedelta(days=(total_months*31))
            receipt.auto_renew = True
        elif term_type == TermType.MONTHLY_SUBSCRIPTION:
            total_months = 1
            receipt.end_date = timezone.now() + timedelta(days=(total_months*31))
            receipt.auto_renew = True
        elif term_type == TermType.QUARTERLY_SUBSCRIPTION:
            total_months = 3
            receipt.end_date = timezone.now() + timedelta(days=(total_months*31))
            receipt.auto_renew = True
        elif term_type == TermType.SEMIANNUAL_SUBSCRIPTION:
            total_months = 6
            receipt.end_date = timezone.now() + timedelta(days=(total_months*31))
            receipt.auto_renew = True
        elif term_type == TermType.ANNUAL_SUBSCRIPTION:
            total_months = 12
            receipt.end_date = timezone.now() + timedelta(days=(total_months*31))
            receipt.auto_renew = True
        elif term_type == TermType.PERPETUAL:
            receipt.auto_renew = False
        elif term_type == TermType.ONE_TIME_USE:
            receipt.auto_renew = False
        return receipt

    def create_receipts(self):
        if self.payment.success and self.invoice.status == Invoice.InvoiceStatus.COMPLETE:
            for order_item in self.invoice.order_items.all():
                for product in order_item.offer.products.all():
                    receipt = self.create_receipt_by_term_type(product, order_item, order_item.offer.terms)
                    receipt.save()
                    receipt.products.add(product)

    def update_subscription_receipt(self, subscription, subscription_id, status):
        """
        subscription: OrderItem
        subscription_id: int
        status: PurchaseStatus
        """
        subscription_receipt = self.invoice.order_items.get(offer=subscription.offer).receipts.get(transaction=self.payment.transaction)
        subscription_receipt.meta['subscription_id'] = subscription_id
        subscription_receipt.status = status
        subscription_receipt.save()

    def amount(self):   # Retrieves the total amount from the invoice
        self.invoice.update_totals()
        return self.invoice.total

    def amount_without_subscriptions(self):
        subscription_total = sum([ oi.total for oi in self.invoice.order_items.filter(offer__terms=TermType.SUBSCRIPTION)])

        amount = self.invoice.total - subscription_total
        return amount

    def get_transaction_id(self):
        return "{}-{}-{}-{}".format(self.invoice.profile.pk, settings.SITE_ID, self.invoice.pk, str(self.payment.created)[-12:-6])

    def get_billing_address_form_data(self, form_data, form_class):
        self.billing_address = form_class(form_data)
    
    def get_payment_info_form_data(self, form_data, form_class):
        self.payment_info = form_class(form_data)

    def is_data_valid(self):
        if not (self.billing_address.is_valid() and self.payment_info.is_valid() and self.invoice and self.invoice.order_items.count()):
            return False
        return True
    #-------------------
    # Data for the View

    def get_checkout_context(self, request=None, context={}):
        '''
        The Invoice plus any additional values to include in the payment record.
        '''
        # context = deepcopy(context)
        context['invoice'] = self.invoice
        return context

    def get_header_javascript(self):
        """
        Scripts that are expected to show in the top of the template.

        This will return a list of relative static URLs to the scripts.
        """
        return []

    def get_javascript(self):
        """
        Scripts added to the bottom of the page in the normal js location.

        This will return a list of relative static URLs to the scripts.
        """
        return []

    def get_template(self):
        """
        Unique partial template for the processor
        """
        pass

    #-------------------
    # Process a Payment

    def authorize_payment(self):
        """
        This runs the chain of events in a transaction.
        
        This should not be overriden.  Override one of the methods it calls if you need to.
        """
        self.status = PurchaseStatus.QUEUED     # TODO: Set the status on the invoice.  Processor status should be the invoice's status.
        vendor_pre_authorization.send(sender=self.__class__, invoice=self.invoice)
        self.pre_authorization()

        self.status = PurchaseStatus.ACTIVE     # TODO: Set the status on the invoice.  Processor status should be the invoice's status.
        vendor_process_payment.send(sender=self.__class__, invoice=self.invoice)
        if not self.invoice.total:
            self.free_payment()
        elif self.is_data_valid():
            self.process_payment()
        else:
            return None

        vendor_post_authorization.send(sender=self.__class__, invoice=self.invoice)
        self.post_authorization()

        #TODO: Set the status based on the result from the process_payment()

    def pre_authorization(self):
        """
        Called before the authorization begins.
        """
        pass

    def process_payment(self):
        """
        Called to handle the authorization.  
        This is where the core of the payment processing happens.
        """
        # Gateway Transaction goes here...
        self.transaction_submitted = True
        pass
            
    def free_payment(self):
        """
        Called to handle an invoice with total zero.  
        This are the base internal steps to process a free payment.
        """
        self.transaction_submitted = True
        self.create_payment_model()

        self.payment.success = True
        self.payment.transation = f"{self.payment.pk}-free"
        self.payment.payee_full_name = " ".join([self.invoice.profile.user.first_name, self.invoice.profile.user.last_name])
        self.payment.save()
        
        self.update_invoice_status(Invoice.InvoiceStatus.COMPLETE)

        self.create_receipts()

    def post_authorization(self):
        """
        Called after the authorization is complete.
        """
        pass

    def capture_payment(self):
        """
        Called to handle the capture.  (some gateways handle this at the same time as authorize_payment() )
        """
        pass

    #-------------------
    # Process a Subscription
    
    def subscription_payment(self):
        pass

    def update_subscription_payment(self):
        pass

    def cancel_subscription_payment(self):
        pass

    #-------------------
    # Refund a Payment

    def refund_payment(self):
        pass

    


from django.core.exceptions import ImproperlyConfigured
from django.contrib import messages
from django.utils.translation import ugettext as _
from django.http import Http404
from django.contrib.sites.models import Site
from django.shortcuts import redirect


class ProductRequiredMixin():
    """
    Checks to see if a user has a required product and if not, redirects them.
    """

    product_queryset = None
    product_model = None
    product_redirect = "/"
    product_owned = False

    def dispatch(self, request, *args, **kwargs):
        """
        Verify that the current user owns the product.  Checks on all HTTP methods.
        """
        
        if not self.user_has_product():
            print("No Product Ownership")
            return self.handle_no_product()
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['product_owned'] = self.product_owned        # Variable set by ProductRequiredMixin
        return context

    def user_has_product(self):
        """
        Check to see if a user has a viable product license based on the get_product_queryset() method.

        TODO: move this to some kind of caching or session variable to avoid hitting the DB on every request.
        """

        if self.request.user.is_anonymous:
            self.product_owned = False
        else:
            products = self.get_product_queryset()
            self.product_owned = self.request.user.customer_profile.filter(site=Site.objects.get_current()).get().has_product(products)

        return self.product_owned

    def get_product_queryset(self):
        """
        Method to get the Product(s) needed for the check.  Can be overridden to handle complex queries.
        """

        if self.product_queryset is None:
            if self.product_model:
                return self.product_model.on_site.all()    # Only provide list of products on the current site.
            else:
                raise ImproperlyConfigured(
                    "%(cls)s is missing a Product QuerySet. Define "
                    "%(cls)s.product_model, %(cls)s.product_queryset, or override "
                    "%(cls)s.get_product_queryset()." % {
                        'cls': self.__class__.__name__
                    }
                )
        return self.product_queryset.all()

    def handle_no_product(self):
        '''
        The result to be returned if no viable product is found owned by the user.
        If None is returned, default to a 404 response.

        This method allows for a dynamic redirect to be used.
        Default is to use the product_redirect path.
        '''

        if not self.product_redirect:
            raise Http404(_("No %(verbose_name)s found matching the query") %
                    {'verbose_name': self.get_product_queryset().model._meta.verbose_name})

        messages.info(self.request, _("Product Purchase required."))
        return redirect(self.product_redirect)



from django.conf import settings
from django.contrib.sites.models import Site
from django.contrib.sites.managers import CurrentSiteManager
from django.db import models
from django.db.models import Q, QuerySet
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from .base import CreateUpdateModelBase
from .choice import CURRENCY_CHOICES, TermType
from .invoice import Invoice
from .utils import set_default_site_id
from vendor.config import DEFAULT_CURRENCY

#####################
# CUSTOMER PROFILE
#####################


class CustomerProfile(CreateUpdateModelBase):
    '''
    Additional customer information related to purchasing.
    This is what the Invoices are attached to.  This is abstracted from the user model directly do it can be mre flexible in the future.
    '''
    user = models.ForeignKey(settings.AUTH_USER_MODEL, verbose_name=_("User"), null=True, on_delete=models.SET_NULL, related_name="customer_profile")
    currency = models.CharField(_("Currency"), max_length=4, choices=CURRENCY_CHOICES,default=DEFAULT_CURRENCY)      # User's default currency
    site = models.ForeignKey(Site, verbose_name=_("Site"), on_delete=models.CASCADE, default=set_default_site_id, related_name="customer_profile")                      # For multi-site support

    objects = models.Manager()
    on_site = CurrentSiteManager()

    class Meta:
        verbose_name = "Customer Profile"
        verbose_name_plural = "Customer Profiles"

    def __str__(self):
        if not self.user:
            return "New Customer Profile"
        return "{username} Customer Profile".format(username=self.user.username)

    def get_customer_profile_display(self):
        return str(self.user.username) + _("Customer Profile")

    def revert_invoice_to_cart(self):
        cart = self.invoices.get(status=Invoice.InvoiceStatus.CHECKOUT)
        cart.status = Invoice.InvoiceStatus.CART
        cart.save()

    def get_cart(self):
        if self.has_invoice_in_checkout():
            self.revert_invoice_to_cart()
        cart, created = self.invoices.get_or_create(status=Invoice.InvoiceStatus.CART)
        return cart

    def get_checkout_cart(self):
        return self.invoices.filter(status=Invoice.InvoiceStatus.CHECKOUT).first()

    def get_cart_or_checkout_cart(self):
        carts_status = [cart.status for cart in self.invoices.filter(status__in=[Invoice.InvoiceStatus.CHECKOUT, Invoice.InvoiceStatus.CART])]
        
        if Invoice.InvoiceStatus.CHECKOUT in carts_status:
            return self.invoices.get(status=Invoice.InvoiceStatus.CHECKOUT)
        else:
            cart, created = self.invoices.get_or_create(status=Invoice.InvoiceStatus.CART)
            return cart
    
    def has_invoice_in_checkout(self):
        return bool(self.invoices.filter(status=Invoice.InvoiceStatus.CHECKOUT).count())
        
    def filter_products(self, products):
        """
        returns the list of reciepts that the user has a reciept for filtered by the products provided.
        """
        now = timezone.now()

        # Queryset or List of model records
        if isinstance(products, QuerySet) or isinstance(products, list):
            return self.receipts.filter(Q(products__in=products),
                                Q(start_date__lte=now) | Q(start_date=None),
                                Q(end_date__gte=now) | Q(end_date=None))

        # Single model record
        return self.receipts.filter(Q(products=products),
                                Q(start_date__lte=now) | Q(start_date=None),
                                Q(end_date__gte=now) | Q(end_date=None))

    def has_product(self, products):
        """
        returns true/false if the user has a receipt to a given product(s)
        it also checks against elegibility start/end/empty dates on consumable products and subscriptions
        """
        return bool(self.filter_products(products).count())

    def get_cart_items_count(self):
        cart = self.get_cart_or_checkout_cart()

        return cart.order_items.all().count()



import csv

from django.http import StreamingHttpResponse
from django.utils.timezone import localtime
from django.contrib.sites.models import Site
# from django.shortcuts import render, redirect
# from django.contrib import messages
# from django.contrib.auth.mixins import LoginRequiredMixin
# from django.urls import reverse
# from django.conf import settings
# from django.utils.translation import ugettext as _
# from django.core.exceptions import ObjectDoesNotExist
from django.contrib.auth.mixins import PermissionRequiredMixin

# from django.views.generic.edit import DeleteView
from django.views.generic.list import BaseListView
# from django.views.generic.detail import DetailView
# from django.views.generic import TemplateView

from vendor.models import Receipt, Invoice

# from vendor.models import Offer, Invoice, Payment, Address, CustomerProfile
# from vendor.models.choice import TermType
# from vendor.models.utils import set_default_site_id
# from vendor.processors import PaymentProcessor
# from vendor.forms import BillingAddressForm, CreditCardForm, AccountInformationForm

# # from vendor.models.address import Address as GoogleAddress

# # The Payment Processor configured in settings.py
# payment_processor = PaymentProcessor


# class CartView(LoginRequiredMixin, DetailView):
#     '''
#     View items in the cart
#     '''
#     model = Invoice

#     def get_object(self):
#         profile, created = self.request.user.customer_profile.get_or_create(
#             site=set_default_site_id())
#         return profile.get_cart()


class Echo:
    """An object that implements just the write method of the file-like
    interface.
    """
    
    def write(self, value):
        """Write the value by returning it, instead of storing in a buffer."""
        return value


# class RecieptListCSV(PermissionRequiredMixin, BaseListView):
class CSVStreamRowView(BaseListView):
    """A base view for displaying a list of objects."""

    filename = "reciept_list.csv"
    # headers = 

    def get_queryset(self):
        return super().get_queryset()
    
    def get_row_data(self):
        rows = (["Row {}".format(idx), str(idx)] for idx in range(500))
        return rows

    def get(self, request, *args, **kwargs):
        rows = self.get_row_data()
        pseudo_buffer = Echo()
        writer = csv.writer(pseudo_buffer)
        #TODO: Figure out what to prepend the header without breaking the streaming generator
        response = StreamingHttpResponse((writer.writerow(row) for row in rows), content_type="text/csv")

        # Set the filename
        response['Content-Disposition'] = 'attachment; filename="{}"'.format(self.filename)
        return response


class RecieptListCSV(CSVStreamRowView):
    filename = "reciepts.csv"
    model = Receipt

    def get_queryset(self):
        # TODO: Update to handle ranges from a POST
        return self.model.objects.filter(profile__site=Site.objects.get_current())      # Return reciepts only for profiles on this site

    def get_row_data(self):
        object_list = self.get_queryset()
        # header = ["RECIEPT_ID", "CREATED_TIME(ISO)", "USERNAME", "INVOICE_ID", "ORDER_ITEM", "OFFER_ID", "QUANTITY", "TRANSACTION_ID", "STATUS"]
        rows = [[str(obj.pk), obj.created.isoformat(), obj.profile.user.username, obj.order_item.invoice.pk, obj.order_item.offer, obj.order_item.pk, obj.order_item.quantity, obj.transaction, obj.get_status_display()] for obj in object_list]
        return rows

 
class InvoiceListCSV(CSVStreamRowView):
    filename = "invoices.csv"
    model = Invoice

    def get_queryset(self):
        # TODO: Update to handle ranges from a POST
        return self.model.on_site.all()
    
    def get_row_data(self):
        object_list = self.get_queryset()
        # header = ["INVOICE_ID", "CREATED_TIME(ISO)", "USERNAME", "CURRENCY", "TOTAL"]
        rows = ([str(obj.pk), obj.created.isoformat(), str(obj.profile.user.username), obj.currency, obj.total] for obj in object_list)
        return rows



import uuid

from autoslug import AutoSlugField
from decimal import Decimal, ROUND_UP
from django.conf import settings
from django.contrib.sites.models import Site
from django.contrib.sites.managers import CurrentSiteManager
from django.core.exceptions import FieldError
from django.db import models
from django.db.models import Q
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from iso4217 import Currency

from vendor.config import VENDOR_PRODUCT_MODEL, DEFAULT_CURRENCY

from .base import CreateUpdateModelBase
from .choice import TermType
from .utils import set_default_site_id
#########
# OFFER
#########

class Offer(CreateUpdateModelBase):
    '''
    Offer attaches to a record from the designated VENDOR_PRODUCT_MODEL.  
    This is so more than one offer can be made per product, with different 
    priorities.  It also allows for the bundling of several products into
    a single Offer on the site.
    '''
    uuid = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)                                # Used to track the product
    slug = AutoSlugField(populate_from='name', unique_with='site__id')                                               # SEO friendly 
    site = models.ForeignKey(Site, verbose_name=_("Site"), on_delete=models.CASCADE, default=set_default_site_id, related_name="product_offers")                      # For multi-site support
    name = models.CharField(_("Name"), max_length=80, blank=True)                                           # If there is only a Product and this is blank, the product's name will be used, oterhwise it will default to "Bundle: <product>, <product>""
    start_date = models.DateTimeField(_("Start Date"), help_text=_("What date should this offer become available?"))
    end_date = models.DateTimeField(_("End Date"), blank=True, null=True, help_text=_("Expiration Date?"))
    terms =  models.IntegerField(_("Terms"), default=0, choices=TermType.choices)
    term_details = models.JSONField(_("Term Details"), default=dict, blank=True, null=True)
    term_start_date = models.DateTimeField(_("Term Start Date"), help_text=_("When is this product available to use?"), blank=True, null=True) # Useful for Event Tickets or Pre-Orders
    available = models.BooleanField(_("Available"), default=False, help_text=_("Is this currently available?"))
    bundle = models.BooleanField(_("Is a Bundle?"), default=False, help_text=_("Is this a product bundle? (auto-generated)"))  # Auto-generated based on if the count of the products is greater than 1.
    offer_description = models.TextField(_("Offer Description"), blank=True, null=True)
    list_bundle_items = models.BooleanField(_("List Bundled Items"), default=False, help_text=_("When showing to customers, display the included items in a list?"))
    allow_multiple = models.BooleanField(_("Allow Multiple Purchase"), default=False, help_text=_("Confirm the user wants to buy multiples of the product where typically there is just one purchased at a time."))

    objects = models.Manager()
    on_site = CurrentSiteManager()

    class Meta:
        verbose_name = "Offer"
        verbose_name_plural = "Offers"

    def __str__(self):
        return self.name

    def get_msrp(self, currency=DEFAULT_CURRENCY):
        return sum([p.get_msrp(currency) for p in self.products.all()])

    def current_price(self, currency=DEFAULT_CURRENCY):
        '''
        Finds the highest priority active price and returns that, otherwise returns msrp total.
        '''
        now = timezone.now()
        price = self.prices.filter( Q(start_date__lte=now) | Q(start_date=None),
                                    Q(end_date__gte=now) | Q(end_date=None)
                                    ).order_by('-priority').first()            # first()/last() returns the model object or None

        if price:
            result = price.cost
        else:
            result = self.get_msrp(currency)                            # If there is no price for the offer, all MSRPs should be summed up for the "price". 

        return result

    def add_to_cart_link(self):
        return reverse("vendor:add-to-cart", kwargs={"slug":self.slug})

    def remove_from_cart_link(self):
        return reverse("vendor:remove-from-cart", kwargs={"slug":self.slug})
    
    def set_name_if_empty(self):
        product_names = [ product.name for product in self.products.all() ]
        if len(product_names) == 1:
            self.name = product_names[0]
        else:
            self.name = "Bundle: " + ", ".join(product_names)

    @property
    def description(self):
        if self.offer_description:
            return self.offer_description
        else:
            return self.products.all().first().description
    
    def savings(self, currency=DEFAULT_CURRENCY):
        savings = self.get_msrp(currency) - self.current_price()
        if savings < 0:
            return Decimal(0).quantize(Decimal('.00'), rounding=ROUND_UP)
        return Decimal(savings).quantize(Decimal('.00'), rounding=ROUND_UP)



# Generated by Django 3.1 on 2020-09-25 19:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0003_auto_20200915_1839'),
    ]

    operations = [
        migrations.AddField(
            model_name='offer',
            name='offer_description',
            field=models.TextField(blank=True, null=True, verbose_name='Offer Description'),
        ),
    ]



from django.apps import apps
from django.contrib.auth.mixins import LoginRequiredMixin
from django.shortcuts import redirect, render
from django.urls import reverse
from django.views.generic import TemplateView
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, UpdateView
from django.views.generic.list import ListView
from vendor.config import VENDOR_PRODUCT_MODEL
from vendor.models import Invoice, Offer, Price
from vendor.forms import ProductForm, OfferForm, PriceForm, PriceFormSet

Product = apps.get_model(VENDOR_PRODUCT_MODEL)
#############
# Admin Views

class AdminDashboardView(LoginRequiredMixin, ListView):
    '''
    List of the most recent invoices generated on the current site.
    '''
    template_name = "vendor/manage/dashboard.html"
    model = Invoice

    def get_queryset(self):
        return self.model.on_site.all()[:10]    # Return the most recent 10


class AdminInvoiceListView(LoginRequiredMixin, ListView):
    '''
    List of all the invoices generated on the current site.
    '''
    template_name = "vendor/manage/invoice_list.html"
    model = Invoice

    def get_queryset(self):
        return self.model.on_site.filter(status__gt=Invoice.InvoiceStatus.CART)  # ignore cart state invoices


class AdminInvoiceDetailView(LoginRequiredMixin, DetailView):
    '''
    Details of an invoice generated on the current site.
    '''
    template_name = "vendor/manage/invoice_detail.html"
    model = Invoice
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'


class AdminProductListView(LoginRequiredMixin, ListView):
    '''
    Creates a Product to be added to offers
    '''
    template_name = "vendor/manage/products.html"
    model = Product
    queryset = Product.on_site.all()


class AdminProductUpdateView(LoginRequiredMixin, UpdateView):
    '''
    Details of an invoice generated on the current site.
    '''
    template_name = "vendor/manage/product.html"
    model = Product
    form_class = ProductForm
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'

    def form_valid(self, form):
        product = form.save(commit=False)
        product.save()
        return redirect('vendor_admin:manager-product-list')


class AdminProductCreateView(LoginRequiredMixin, CreateView):
    '''
    Creates a Product to be added to offers
    '''
    template_name = "vendor/manage/product.html"
    form_class = ProductForm

    def form_valid(self, form):
        product = form.save(commit=False)
        product.save()
        return redirect('vendor_admin:manager-product-list')

class AdminOfferListView(LoginRequiredMixin, ListView):
    '''
    Creates a Product to be added to offers
    '''
    template_name = "vendor/manage/offers.html"
    model = Offer
    queryset = Offer.on_site.all()


class AdminOfferUpdateView(LoginRequiredMixin, UpdateView):
    '''
    Details of an invoice generated on the current site.
    '''
    template_name = "vendor/manage/offer.html"
    model = Offer
    form_class = OfferForm
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context['formset'] = PriceFormSet(instance=Offer.objects.get(uuid=context['view'].kwargs['uuid']))

        return context


    def form_valid(self, form):
        price_formset = PriceFormSet(self.request.POST, self.request.FILES, instance=Offer.objects.get(uuid=self.kwargs['uuid']))

        if not (price_formset.is_valid() or form.is_valid()):
            return render(self.request, self.template_name, {'form': form, 'formsert': price_formset})

        offer = form.save(commit=False)

        if len(form.cleaned_data['products']):
            offer.bundle=True
        
        offer.save()

        for product in form.cleaned_data['products']:
            offer.products.add(product)

        for price_form in price_formset:
            price = price_form.save(commit=False)
            price.offer = offer
            price.save()

        return redirect('vendor_admin:manager-offer-list')


class AdminOfferCreateView(LoginRequiredMixin, CreateView):
    '''
    Creates a Product to be added to offers
    '''
    template_name = "vendor/manage/offer.html"
    form_class = OfferForm    

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context['formset'] = PriceFormSet()

        return context

    def post(self, request):
        offer_form = self.form_class(request.POST)
        price_formset = PriceFormSet(request.POST)

        if not (price_formset.is_valid() and offer_form.is_valid()):
            return render(request, self.template_name, {'form': offer_form, 'formsert': price_formset})

        offer = offer_form.save(commit=False)
        if len(offer_form.cleaned_data['products']):
            offer.bundle=True

        offer.save()
        for product in offer_form.cleaned_data['products']:
            offer.products.add(product)


        for price_form in price_formset:
            price = price_form.save(commit=False)
            price.offer = offer
            price.save()

        return redirect('vendor_admin:manager-offer-list')


"""
Payment processor for Authorize.net.
"""
import ast
from datetime import datetime
from decimal import Decimal, ROUND_DOWN

from django.conf import settings

from vendor.config import VENDOR_PAYMENT_PROCESSOR

try:
    from authorizenet import apicontractsv1
    from authorizenet.apicontrollers import *
except ModuleNotFoundError:
    if VENDOR_PAYMENT_PROCESSOR == "authorizenet.AuthorizeNetProcessor":
        print("WARNING: authorizenet module not found.  Install the library if you want to use the AuthorizeNetProcessor.")
        raise
    pass

from vendor.forms import CreditCardForm, BillingAddressForm
from vendor.models.choice import TransactionTypes, PaymentTypes, TermType, PurchaseStatus
from vendor.models.invoice import Invoice
from vendor.models.address import Country
from .base import PaymentProcessorBase


class AuthorizeNetProcessor(PaymentProcessorBase):
    """
    Implementation of Authoirze.Net SDK
    https://apitest.authorize.net/xml/v1/schema/AnetApiSchema.xsd
    """

    AUTHORIZE = "authOnlyTransaction"
    AUTHORIZE_CAPTURE = "authCaptureTransaction"
    CAPTURE = "captureOnlyTransaction"
    REFUND = "refundTransaction"
    PRIOR_AUTHORIZE_CAPTURE = "priorAuthCaptureTransaction"
    VOID = "voidTransaction"
    GET_DETAILS = "getDetailsTransaction"
    AUTHORIZE_CONTINUE = "authOnlyContinueTransaction"
    AUTHORIZE_CAPTURE_CONTINUE = "authCaptureContinueTransaction"

    GET_SETTLED_BATCH_LIST = "getSettledBatchListRequest"

    """
    Controller executes the transaction, that process a transaction type
    """
    controller = None
    transaction = None
    merchant_auth = None
    transaction_type = None

    def __str__(self):
        return 'Authorize.Net'

    def get_checkout_context(self, request=None, context={}):
        context = super().get_checkout_context(context=context)
        if 'credit_card_form' not in context:
            context['credit_card_form'] = CreditCardForm(initial={'payment_type': PaymentTypes.CREDIT_CARD})
        if 'billing_address_form' not in context:
            context['billing_address_form'] = BillingAddressForm()
        return context

    def processor_setup(self):
        """
        Merchant Information needed to aprove the transaction. 
        """
        if not (settings.AUTHORIZE_NET_TRANSACTION_KEY and settings.AUTHORIZE_NET_API_ID):
            raise ValueError(
                "Missing Authorize.net keys in settings: AUTHORIZE_NET_TRANSACTION_KEY and/or AUTHORIZE_NET_API_ID")
        self.merchant_auth = apicontractsv1.merchantAuthenticationType()
        self.merchant_auth.transactionKey = settings.AUTHORIZE_NET_TRANSACTION_KEY
        self.merchant_auth.name = settings.AUTHORIZE_NET_API_ID
        self.init_payment_type_switch()
        self.init_transaction_types()

    def init_transaction_types(self):
        self.transaction_types = {
            TransactionTypes.AUTHORIZE: self.AUTHORIZE,
            TransactionTypes.CAPTURE: self.CAPTURE,
            TransactionTypes.REFUND: self.REFUND,
        }

    def init_payment_type_switch(self):
        """
        Initializes the Payment Types create functions
        """
        self.payment_type_switch = {
            PaymentTypes.CREDIT_CARD: self.create_credit_card_payment,
            PaymentTypes.BANK_ACCOUNT: self.create_bank_account_payment,
            PaymentTypes.PAY_PAL: self.create_pay_pal_payment,
            PaymentTypes.MOBILE: self.create_mobile_payment,
        }

    ##########
    # Authorize.net Object creations
    ##########

    def create_transaction(self):
        """
        This creates the main transaction to be processed.
        """
        transaction = apicontractsv1.createTransactionRequest()
        transaction.merchantAuthentication = self.merchant_auth
        transaction.refId = self.get_transaction_id()
        return transaction

    def create_transaction_type(self, trans_type):
        """
        Creates the transaction type instance with the amount
        """
        transaction_type = apicontractsv1.transactionRequestType()
        transaction_type.transactionType = trans_type
        transaction_type.currencyCode = 'USD'
        return transaction_type

    def create_credit_card_payment(self):
        """
        Creates and credit card payment type instance form the payment information set. 
        """
        creditCard = apicontractsv1.creditCardType()
        creditCard.cardNumber = self.payment_info.data.get('card_number')
        creditCard.expirationDate = "-".join([self.payment_info.data.get('expire_year'), self.payment_info.data.get('expire_month')])
        creditCard.cardCode = str(self.payment_info.data.get('cvv_number'))
        return creditCard

    def create_bank_account_payment(self):
        raise NotImplementedError

    def create_pay_pal_payment(self):
        raise NotImplementedError

    def create_mobile_payment(self):
        raise NotImplementedError

    def create_authorize_payment(self):
        """
        Creates a payment instance acording to the billing information captured
        """
        payment = apicontractsv1.paymentType()
        payment.creditCard = self.payment_type_switch[int(
            self.payment_info.data.get('payment_type'))]()
        return payment

    def create_customer_data(self):
        customerData = apicontractsv1.customerDataType()
        customerData.type = "individual"
        customerData.id = str(self.invoice.profile.user.pk)
        customerData.email = self.invoice.profile.user.email
        return customerData

    def set_transaction_request_settings(self, settings):
        raise NotImplementedError

    def set_transaction_request_client_ip(self, client_ip):
        raise NotImplementedError

    def create_line_item(self, order_item):
        """
        Create a single line order_item (products) that is attached to a line order_item array
        """
        line_item = apicontractsv1.lineItemType()
        line_item.itemId = str(order_item.pk)
        line_item.name = order_item.name[:30]
        line_item.description = str(order_item.offer.description)[:250]
        line_item.quantity = str(order_item.quantity)
        line_item.unitPrice = str(order_item.price)
        return line_item

    def create_line_item_array(self, order_items):
        """
        Creates a list o line order_items(products) that are attached to the transaction type
        """
        line_items = apicontractsv1.ArrayOfLineItem()
        for order_item in order_items:
            line_items.lineItem.append(self.create_line_item(order_item))
        return line_items

    def create_tax(self, tax):
        raise NotImplementedError

    def create_shipping(self, shipping):
        raise NotImplementedError

    def create_billing_address(self):
        """
        Creates Billing address to improve security in transaction
        """
        billing_address = apicontractsv1.customerAddressType()
        billing_address.firstName = " ".join(self.payment_info.data.get('full_name', "").split(" ")[:-1])[:50]
        billing_address.lastName = (self.payment_info.data.get('full_name', "").split(" ")[-1])[:50]
        billing_address.company = self.billing_address.data.get('company', "")[:50]
        billing_address.address = ", ".join([self.billing_address.data.get('address_1',""), str(self.billing_address.data.get('address_2', ""))])[:60]
        billing_address.city = self.billing_address.data.get("city", "")[:40]
        billing_address.state = self.billing_address.data.get("state", "")[:40]
        billing_address.zip = self.billing_address.data.get("postal_code")[:20]
        country = Country(int(self.billing_address.data.get("country")))
        billing_address.country = str(country.name)
        return billing_address

    def create_customer(self):
        raise NotImplementedError

    def get_payment_occurrences(self, subscription, subscription_type):
        if subscription_type == TermType.SUBSCRIPTION:
            return subscription.offer.term_details['payment_occurrences']
        elif subscription_type == TermType.MONTHLY_SUBSCRIPTION:
            return 1
        elif subscription_type == TermType.QUARTERLY_SUBSCRIPTION:
            return 1
        elif subscription_type == TermType.SEMIANNUAL_SUBSCRIPTION:
            return 1
        elif subscription_type == TermType.ANNUAL_SUBSCRIPTION:
            return 1

    def get_period_length(self, subscription, subscription_type):
        if subscription_type == TermType.SUBSCRIPTION:
            return subscription.offer.term_details['period_length']
        elif subscription_type == TermType.MONTHLY_SUBSCRIPTION:
            return 1
        elif subscription_type == TermType.QUARTERLY_SUBSCRIPTION:
            return 4
        elif subscription_type == TermType.SEMIANNUAL_SUBSCRIPTION:
            return 6
        elif subscription_type == TermType.ANNUAL_SUBSCRIPTION:
            return 12

    def get_trail_occurrences(self, subscription):
        return subscription.offer.term_details.get('trial_occurrences', 0)

    def create_payment_scheduale_interval_type(self, subscription, subscription_type):
        """
        Create an interval schedule with fixed months.
        period_length: The period length the service paided mor last 
            eg: period_length = 2. the user will be billed every 2 months.
        payment_occurrences: The number of occurrences the payment should be made.
            eg: payment_occurrences = 6. There will be six payments made at each period_length.
        trial_occurrences: The number of ignored payments out of the payment_occurrences
        """
        trial_occurrences = self.get_trail_occurrences(subscription)

        payment_schedule = apicontractsv1.paymentScheduleType()
        payment_schedule.interval = apicontractsv1.paymentScheduleTypeInterval()
        payment_schedule.interval.unit = apicontractsv1.ARBSubscriptionUnitEnum.months
        payment_schedule.startDate = datetime.now()

        payment_schedule.interval.length = self.get_period_length(subscription, subscription_type)
        payment_schedule.totalOccurrences = self.get_payment_occurrences(subscription, subscription_type)
        payment_schedule.trialOccurrences = trial_occurrences

        return payment_schedule

    ##########
    # Django-Vendor to Authoriaze.net data exchange functions
    ##########
    def save_payment_transaction(self):
        self.payment.success = self.transaction_submitted
        self.payment.transaction = self.transaction_response.transId.text
        response = self.transaction_response.__dict__
        if 'errors' in response:
            response.pop('errors')
        if 'messages' in response:
            response.pop('messages')

        self.payment.result = {'raw': str({**self.transaction_message, **response})}
        self.payment.result['account_number'] = ast.literal_eval(self.payment.result['raw']).get('accountNumber')[-4:]
        self.payment.result['account_type'] = ast.literal_eval(self.payment.result['raw']).get('accountType')

        self.payment.payee_full_name = self.payment_info.data.get(
            'full_name')
        self.payment.payee_company = self.billing_address.data.get(
            'company')
        billing_address = self.billing_address.save(commit=False)
        billing_address.profile = self.invoice.profile
        billing_address.save()
        self.payment.billing_address = billing_address
        self.payment.save()

    def check_response(self, response):
        """
        Checks the transaction response and set the transaction_submitted and transaction_response variables
        """
        self.transaction_response = response.transactionResponse
        self.transaction_message = {}
        self.transaction_submitted = False
        self.transaction_message['msg'] = ""
        if response is not None:
            # Check to see if the API request was successfully received and acted upon
            if response.messages.resultCode == "Ok":
                # Since the API request was successful, look for a transaction response
                # and parse it to display the results of authorizing the card
                if hasattr(response.transactionResponse, 'messages') is True:
                    self.transaction_submitted = True
                    self.transaction_message['msg'] = "Payment Complete"
                    self.transaction_message['trans_id'] = response.transactionResponse.transId
                    self.transaction_message['response_code'] = response.transactionResponse.responseCode
                    self.transaction_message['code'] = response.transactionResponse.messages.message[0].code
                    self.transaction_message['message'] = response.transactionResponse.messages.message[0].description
                else:
                    self.transaction_message['msg'] = 'Failed Transaction.'
                    if hasattr(response.transactionResponse, 'errors') is True:
                        self.transaction_message['error_code'] = response.transactionResponse.errors.error[0].errorCode
                        self.transaction_message['error_text'] = response.transactionResponse.errors.error[0].errorText
            # Or, print errors if the API request wasn't successful
            else:
                self.transaction_message['msg'] = 'Failed Transaction.'
                if hasattr(response, 'transactionResponse') is True and hasattr(response.transactionResponse, 'errors') is True:
                    self.transaction_message['error_code'] = response.transactionResponse.errors.error[0].errorCode
                    self.transaction_message['error_text'] = response.transactionResponse.errors.error[0].errorText
                else:
                    self.transaction_message['error_code'] = response.messages.message[0]['code'].text
                    self.transaction_message['error_text'] = response.messages.message[0]['text'].text
        else:
            self.transaction_message['msg'] = 'Null Response.'
    
    def check_subscription_response(self, response):
        self.transaction_response = response
        self.transaction_message = {}
        self.transaction_submitted = False
        self.transaction_message['msg'] = ""
        self.transaction_message['code'] = response.messages.message[0]['code'].text
        self.transaction_message['message'] = response.messages.message[0]['text'].text

        if (response.messages.resultCode=="Ok"):
            self.transaction_submitted = True
            self.transaction_message['msg'] = "Subscription Tansaction Complete"
            if 'subscriptionId' in response.__dict__:
                self.transaction_message['subscription_id'] = response.subscriptionId
        else:
            self.transaction_message['msg'] = "Subscription Tansaction Failed"

    def to_valid_decimal(self, number):
        return Decimal(number).quantize(Decimal('.00'), rounding=ROUND_DOWN)
    ##########
    # Base Processor Transaction Implementations
    ##########
    def process_payment(self):
        self.create_payment_model()

        # Init transaction
        self.transaction = self.create_transaction()
        self.transaction_type = self.create_transaction_type(settings.AUTHOIRZE_NET_TRANSACTION_TYPE_DEFAULT)
        self.transaction_type.amount = self.to_valid_decimal(self.invoice.total)
        self.transaction_type.payment = self.create_authorize_payment()
        self.transaction_type.billTo = self.create_billing_address()

        # Optional items for make it easier to read and use on the Authorize.net portal.
        if self.invoice.order_items:
            self.transaction_type.lineItems = self.create_line_item_array(
                self.invoice.order_items.all())

        # You set the request to the transaction
        self.transaction.transactionRequest = self.transaction_type
        self.controller = createTransactionController(self.transaction)
        self.controller.execute()

        # You execute and get the response
        response = self.controller.getresponse()
        self.check_response(response)

        self.save_payment_transaction()

        self.update_invoice_status(Invoice.InvoiceStatus.COMPLETE)

        self.create_receipts()

    def subscription_payment(self, subscription):
        """
        Creates a subscription for a user. Subscriptions can be monthy or yearly.objects.all()
        """
        
        # Setting billing information
        billto = apicontractsv1.nameAndAddressType()
        billto.firstName = " ".join(self.payment_info.data.get('full_name', "").split(" ")[:-1])[:50]
        billto.lastName = (self.payment_info.data.get('full_name', "").split(" ")[-1])[:50]

        # Setting subscription details
        self.transaction_type = apicontractsv1.ARBSubscriptionType()
        self.transaction_type.name = subscription.offer.name
        self.transaction_type.paymentSchedule = self.create_payment_scheduale_interval_type(subscription, subscription.offer.terms)
        self.transaction_type.amount = self.to_valid_decimal(subscription.total)
        self.transaction_type.trialAmount = Decimal('0.00')
        self.transaction_type.billTo = billto
        self.transaction_type.payment = self.create_authorize_payment()

        # Creating the request
        self.transaction = apicontractsv1.ARBCreateSubscriptionRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.subscription = self.transaction_type

        # Creating and executing the controller
        self.controller = ARBCreateSubscriptionController(self.transaction)
        self.controller.execute()
        # Getting the response
        response = self.controller.getresponse()
        
        self.check_subscription_response(response)

        receipt = subscription.receipts.get(transaction=self.payment.transaction)
        receipt.meta = {'raw': str({**self.transaction_message, **response})}
        
        if self.transaction_submitted:
            receipt.meta['subscription_id'] = self.transaction_response.subscriptionId.pyval
            
        receipt.save()


    def update_subscription_payment(self, subscription_id):

        self.transaction_type = apicontractsv1.ARBSubscriptionType()
        self.transaction_type.payment = self.create_authorize_payment()

        self.transaction = apicontractsv1.ARBUpdateSubscriptionRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.subscriptionId = subscriptionId
        self.transaction.subscription = self.transaction_type

        self.controller = ARBUpdateSubscriptionController(self.transaction)
        self.controller.execute()

        response = self.controller.getresponse()

        if (response.messages.resultCode=="Ok"):
            print ("SUCCESS")
            print ("Message Code : %s" % response.messages.message[0]['code'].text)
            print ("Message text : %s" % response.messages.message[0]['text'].text)
        else:
            print ("ERROR")
            print ("Message Code : %s" % response.messages.message[0]['code'].text)
            print ("Message text : %s" % response.messages.message[0]['text'].text)

        return response

    def cancel_subscription_payment(self, reciept, subscription_id):
        self.transaction = apicontractsv1.ARBCancelSubscriptionRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.subscriptionId = str(subscription_id)

        self.controller = ARBCancelSubscriptionController(self.transaction)
        self.controller.execute()

        response = self.controller.getresponse()

        self.check_subscription_response(response)

        if self.transaction_submitted:
            reciept.status = PurchaseStatus.CANCELED
            reciept.save()


    def refund_payment(self, payment):
        # Init transaction
        self.transaction = self.create_transaction()
        self.transaction_type = self.create_transaction_type(
            self.transaction_types[TransactionTypes.REFUND])
        self.transaction_type.amount = self.to_valid_decimal(payment.amount)
        self.transaction_type.refTransId = payment.transaction

        creditCard = apicontractsv1.creditCardType()
        creditCard.cardNumber = ast.literal_eval(payment.result['raw']).get('accountNumber')[-4:]
        creditCard.expirationDate = "XXXX"

        payment_type = apicontractsv1.paymentType()
        payment_type.creditCard = creditCard
        self.transaction_type.payment = payment_type

        self.transaction.transactionRequest = self.transaction_type
        self.controller = createTransactionController(self.transaction)
        self.controller.execute()

        response = self.controller.getresponse()
        self.check_response(response)

        if self.transaction_submitted:
            self.update_invoice_status(Invoice.InvoiceStatus.REFUNDED)

    ##########
    # Reporting API, for transaction retrieval information
    ##########
    def get_settled_batch_list(self, start_date, end_date):
        """
        Gets a list of batches for settled transaction between the start and end date.
        """
        self.transaction = apicontractsv1.getSettledBatchListRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.firstSettlementDate = start_date
        self.transaction.lastSettlementDate = end_date

        self.controller = getSettledBatchListController(self.transaction)
        self.controller.execute()

        response = self.controller.getresponse()

        if response.messages.resultCode == apicontractsv1.messageTypeEnum.Ok and hasattr(response, 'batchList'):
            return [batch for batch in response.batchList.batch]

    def get_transaction_batch_list(self, batch_id):
        """
        Gets the list of settled transaction in a batch. There are sorting and paging option
        that are not currently implemented. It will get the last 1k transactions.
        """
        self.transaction = apicontractsv1.getTransactionListRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.batchId = batch_id

        self.controller = getTransactionListController(self.transaction)
        self.controller.execute()

        response = self.controller.getresponse()

        if response.messages.resultCode == apicontractsv1.messageTypeEnum.Ok and hasattr(response, 'transactions'):
            return [transaction for transaction in response.transactions.transaction]

    def get_transaction_detail(self, transaction_id):
        self.transaction = apicontractsv1.getTransactionDetailsRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.transId = transaction_id

        self.controller = getTransactionDetailsController(self.transaction)
        self.controller.execute()

        response = self.controller.getresponse()

        if response.messages.resultCode == apicontractsv1.messageTypeEnum.Ok:
            return response.transaction

    def get_list_of_subscriptions(self):
        self.transaction = apicontractsv1.ARBGetSubscriptionListRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.searchType = apicontractsv1.ARBGetSubscriptionListSearchTypeEnum.subscriptionActive

        self.controller = ARBGetSubscriptionListController(self.transaction)
        self.controller.execute()

        # Work on the response
        response = self.controller.getresponse()
        if response.messages.resultCode == apicontractsv1.messageTypeEnum.Ok:
            return response.subscriptionDetails.subscriptionDetail





from django.urls import path

from vendor.views import vendor_admin as admin_views
from vendor.views import report as report_views

app_name = "vendor_admin"

urlpatterns = [
    path('', admin_views.AdminDashboardView.as_view(), name="manager-dashboard"),
    path('products/', admin_views.AdminProductListView.as_view(), name="manager-product-list"),
    path('product/<uuid:uuid>/', admin_views.AdminProductUpdateView.as_view(), name="manager-product-update"),
    path('product/', admin_views.AdminProductCreateView.as_view(), name="manager-product-create"),
    path('offers/', admin_views.AdminOfferListView.as_view(), name="manager-offer-list"),
    path('offer/<uuid:uuid>/', admin_views.AdminOfferUpdateView.as_view(), name="manager-offer-update"),
    path('offer/', admin_views.AdminOfferCreateView.as_view(), name="manager-offer-create"),
    path('orders/', admin_views.AdminInvoiceListView.as_view(), name="manager-order-list"),
    path('order/<uuid:uuid>/', admin_views.AdminInvoiceDetailView.as_view(), name="manager-order-detail"),

    # reports
    path('reports/reciepts/download/', report_views.RecieptListCSV.as_view(), name="manager-reciept-download"),
    path('reports/invoices/download/', report_views.InvoiceListCSV.as_view(), name="manager-invoice-download"),
]



from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _

from .choice import CURRENCY_CHOICES
from vendor.config import DEFAULT_CURRENCY

#########
# PRICE
#########

class Price(models.Model):
    offer = models.ForeignKey("vendor.Offer", verbose_name=_("Offer"), on_delete=models.CASCADE, related_name="prices")
    cost = models.FloatField(_("Cost"), blank=True, null=True)
    currency = models.CharField(_("Currency"), max_length=4, choices=CURRENCY_CHOICES, default=DEFAULT_CURRENCY)
    start_date = models.DateTimeField(_("Start Date"), help_text=_("When should the price first become available?"))
    end_date = models.DateTimeField(_("End Date"), blank=True, null=True, help_text=_("When should the price expire?"))
    priority = models.IntegerField(_("Priority"), help_text=_("Higher number takes priority"), blank=True, null=True)

    def save(self, *args, **kwargs):
        if not self.priority:       # TODO: Add check to see if this is the only price on the offer, then let it be 0.  If not, might need to do some assumptions to guess what it should be.
            self.priority = 0

        super().save(*args, **kwargs)

    class Meta:
        verbose_name = "Price"
        verbose_name_plural = "Prices"

    def __str__(self):
        return "{} for {}:{}".format(self.offer.name, dict(CURRENCY_CHOICES)[self.currency], self.cost)



# Generated by Django 3.1.3 on 2020-11-11 07:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0008_offer_allow_multiple'),
    ]

    operations = [
        migrations.AlterField(
            model_name='offer',
            name='terms',
            field=models.IntegerField(choices=[(0, 'Perpetual'), (10, 'Subscription'), (11, 'Monthly Subscription'), (12, 'Quarterly Subscription'), (13, 'Semi-Annual Subscription'), (14, 'Annual Subscription'), (20, 'One-Time Use')], default=0, verbose_name='Terms'),
        ),
    ]



import random
import string

from django.contrib.sites.models import Site
from django.utils.module_loading import import_string

from vendor.config import VENDOR_DATA_ENCODER

encode = import_string('{}.encode'.format(VENDOR_DATA_ENCODER))
decode = import_string('{}.decode'.format(VENDOR_DATA_ENCODER))

###########
# UTILITIES
###########

def random_string(length=8, check=[]):
    letters= string.digits + string.ascii_uppercase
    value = ''.join(random.sample(letters,length))

    if value not in check:
        return value
    return random_string(length=length, check=check)

def generate_sku():
    return random_string()

def set_default_site_id():
    return Site.objects.get_current()



from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _

from .profile import CustomerProfile

#####################
# ADDRESS
#####################

class Country(models.IntegerChoices):
    """
    Following ISO 3166
    https://en.wikipedia.org/wiki/List_of_ISO_3166_country_codes
    """
    
    # AUS = 36, _("Australia")
    # CAN = 124, _("Canada")
    # JPN = 392, _("Japan")
    # MEX = 484, _("Mexico")
    USA = 581, _("United States")

# Can be overridden in the settings.py with a differnt IntegerChoices object
# It should still maintain the ISO-3166 codes for the country numbers and Enumerated keys
COUNTRY_CHOICE = getattr(settings, 'VENDOR_COUNTRY_CHOICE', Country.choices)
COUNTRY_DEFAULT = getattr(settings, 'VENDOR_COUNTRY_DEFAULT', Country.USA)

class Address(models.Model):
    """Address model for use in purchasing.

    Example:
        Format Fields.

        Gartenweg 8         [address1]
                            [address2]
        Rafing              [locality]
        AUSTRIA             [country]
        3741 PULKAU         [postal code]

    Args:
        name (int): What to name the address if they want to identify it for future use.  Default is "Home"
        profile (CustomerProfile): Foreign Key connection to the Customer Profile
        address_1 (str):

    Returns:
        Address(): Returns an instance of the Address model
    """
    name = models.CharField(_("Address Name"), max_length=80, blank=True)                                           # If there is only a Product and this is blank, the product's name will be used, oterhwise it will default to "Bundle: <product>, <product>""
    profile = models.ForeignKey(CustomerProfile, verbose_name=_("Customer Profile"), null=True, on_delete=models.CASCADE, related_name="addresses")
    first_name = models.CharField(_("First Name"), max_length=150, blank=True)
    last_name = models.CharField(_("Last Name"), max_length=150, blank=True)
    address_1 = models.CharField(_("Address 1"), max_length=40, blank=False)
    address_2 = models.CharField(_("Address 2"), max_length=40, blank=True, null=True)
    locality = models.CharField(_("City"), max_length=40, blank=False)
    state = models.CharField(_("State"), max_length=40, blank=False)
    country = models.IntegerField(_("Country"), choices=COUNTRY_CHOICE, default=COUNTRY_DEFAULT)
    postal_code = models.CharField(_("Postal Code"), max_length=16, blank=True)

    # def create_address_from_billing_form(self, billing_form, profile):
    #     locality = Locality()
    #     locality.postal_code = billing_form.data.get('postal_code')
    #     locality.name = billing_form.data.get('postal_code')
    #     locality.state = State.objects.get(pk=billing_form.data.get('state'))
    #     locality.save()

    #     address = Address()
    #     # TODO: regex to only get digits
    #     address.street_number = billing_form.data.get('address_line_1')
    #     address.route = ", ".join([billing_form.data.get('address_line_1', ""),billing_form.data.get('address_line_2', "")])
    #     address.locality = locality
    #     address.raw = ", ".join(
    #         [   billing_form.data.get('name', ""),
    #             billing_form.data.get('address_line_1', ""),
    #             billing_form.data.get('address_line_2', ""), 
    #             billing_form.data.get('city', ""), 
    #             State.objects.get(pk=billing_form.data.get('state', "")).name, 
    #             Country.objects.get(pk=billing_form.data.get('country', "")).name,
    #             billing_form.data.get('postal_code', "")
    #             ])
    #     address.save()
        
    #     self.name = 'Billing: {}'.format(profile.user.username)
    #     self.profile = profile
    #     self.address = address.raw

    def __str__(self):
        return "\n".join([ f"{key}: {value}" for key, value in self.__dict__.items() ])
        
    def get_address(self):
        return f"{self.profile.user}\n{self.address_1}, {self.address_2}\n{self.locality}, {self.state}, {self.postal_code}".replace('None', '')
         
